from kirin import ir, types
from kirin.prelude import basic_no_opt
from kirin.analysis.typeinfer import TypeInference


def test_untable_branch():
    @basic_no_opt
    def unstable(x: int):  # type: ignore
        y = x + 1
        if y > 10:
            z = y
        else:
            z = y + 1.2
        return z

    infer = TypeInference(dialects=unstable.dialects)
    frame, ret = infer.run_no_raise(unstable, types.Int)
    assert ret == types.Union(types.Int, types.Float)

    def stmt_at(block_id, stmt_id) -> ir.Statement:
        return unstable.code.body.blocks[block_id].stmts.at(stmt_id)  # type: ignore

    def results_at(block_id, stmt_id):
        return stmt_at(block_id, stmt_id).results

    assert [frame.entries[result] for result in results_at(0, 0)] == [types.Int]
    assert [frame.entries[result] for result in results_at(0, 1)] == [types.Int]
    assert [frame.entries[result] for result in results_at(0, 2)] == [types.Int]
    assert [frame.entries[result] for result in results_at(0, 3)] == [types.Bool]

    assert [frame.entries[result] for result in results_at(1, 0)] == [types.Int]
    assert [frame.entries[result] for result in results_at(2, 0)] == [types.Float]
    assert [frame.entries[result] for result in results_at(2, 1)] == [types.Float]

    stmt = stmt_at(3, 0)
    assert frame.entries[stmt.args[0]] == (types.Int | types.Float)



from kirin.prelude import basic, basic_no_opt
from kirin.rewrite import Walk, Fixpoint
from kirin.rewrite.cse import CommonSubexpressionElimination


@basic_no_opt
def badprogram(x: int, y: int) -> int:
    a = x + y
    b = x + y
    x = a + b
    y = a + b
    return x + y


def test_cse():
    before = badprogram(1, 2)
    cse = CommonSubexpressionElimination()
    Fixpoint(Walk(cse)).rewrite(badprogram.code)
    after = badprogram(1, 2)

    assert before == after


@basic_no_opt
def cse_constant():
    x = 1
    y = 2
    z = 1
    return x + y + z


def test_cse_constant():
    # NOTE: issue #61
    before = cse_constant()
    cse_constant.print()
    cse = CommonSubexpressionElimination()
    Fixpoint(Walk(cse)).rewrite(cse_constant.code)
    after = cse_constant()
    cse_constant.print()
    assert before == after
    assert len(cse_constant.callable_region.blocks[0].stmts) == 5


def test_cse_constant_int_float():

    @basic(fold=False, typeinfer=True)
    def gv2(x: int):
        y = 1
        z = 1.0
        return y + z + x

    out = gv2(2)

    Walk(CommonSubexpressionElimination()).rewrite(gv2.code)
    gv2.print()

    out2 = gv2(2)

    assert out == out2
    assert type(out) is type(out2)
    assert type(out) is float



"""Type resolution for type inference.

This module contains the type resolution algorithm for type inference.
A simple algorithm is used to resolve the types of the IR by comparing
the input types with the output types.
"""

from dataclasses import field, dataclass

from kirin import types


@dataclass
class TypeResolutionResult:
    """Base class for type resolution results."""

    pass


@dataclass
class ResolutionOk(TypeResolutionResult):
    """Type resolution result for successful resolution."""

    def __bool__(self):
        return True


Ok = ResolutionOk()


@dataclass
class ResolutionError(TypeResolutionResult):
    """Type resolution result for failed resolution."""

    expr: types.TypeAttribute
    value: types.TypeAttribute

    def __bool__(self):
        return False

    def __str__(self):
        return f"expected {self.expr}, got {self.value}"


@dataclass
class TypeResolution:
    """Type resolution algorithm for type inference."""

    vars: dict[types.TypeVar, types.TypeAttribute] = field(default_factory=dict)

    def substitute(self, typ: types.TypeAttribute) -> types.TypeAttribute:
        """Substitute type variables in the type with their values.

        This method substitutes type variables in the given type with their
        values. If the type is a generic type, the method recursively
        substitutes the type variables in the type arguments.

        Args:
            typ: The type to substitute.

        Returns:
            The type with the type variables substituted.
        """
        if isinstance(typ, types.TypeVar):
            return self.vars.get(typ, typ)
        elif isinstance(typ, types.Generic):
            return types.Generic(
                typ.body, *tuple(self.substitute(var) for var in typ.vars)
            )
        elif isinstance(typ, types.Union):
            return types.Union(self.substitute(t) for t in typ.types)
        elif isinstance(typ, types.FunctionType):
            return types.FunctionType(
                params_type=tuple(self.substitute(t) for t in typ.params_type),
                return_type=(
                    self.substitute(typ.return_type) if typ.return_type else None
                ),
            )
        return typ

    def solve(
        self, annot: types.TypeAttribute, value: types.TypeAttribute
    ) -> TypeResolutionResult:
        """Solve the type resolution problem.

        This method compares the expected type `annot` with the actual
        type `value` and returns a result indicating whether the types
        match or not.

        Args:
            annot: The expected type.
            value: The actual type.

        Returns:
            A `TypeResolutionResult` object indicating the result of the
            resolution.
        """
        if isinstance(annot, types.TypeVar):
            return self.solve_TypeVar(annot, value)
        elif isinstance(annot, types.Generic):
            return self.solve_Generic(annot, value)
        elif isinstance(annot, types.Union):
            return self.solve_Union(annot, value)
        elif isinstance(annot, types.FunctionType):
            return self.solve_FunctionType(annot, value)

        if annot.is_subseteq(value):
            return Ok
        else:
            return ResolutionError(annot, value)

    def solve_TypeVar(self, annot: types.TypeVar, value: types.TypeAttribute):
        if annot in self.vars:
            if value.is_subseteq(self.vars[annot]):
                self.vars[annot] = value
            elif self.vars[annot].is_subseteq(value):
                pass
            else:
                return ResolutionError(annot, value)
        else:
            self.vars[annot] = value
        return Ok

    def solve_Generic(self, annot: types.Generic, value: types.TypeAttribute):
        if not isinstance(value, types.Generic):
            return ResolutionError(annot, value)

        if not value.body.is_subseteq(annot.body):
            return ResolutionError(annot.body, value.body)

        for var, val in zip(annot.vars, value.vars):
            result = self.solve(var, val)
            if not result:
                return result

        if not annot.vararg:
            return Ok

        for val in value.vars[len(annot.vars) :]:
            result = self.solve(annot.vararg.typ, val)
            if not result:
                return result
        return Ok

    def solve_FunctionType(self, annot: types.FunctionType, value: types.TypeAttribute):
        if not isinstance(value, types.FunctionType):
            return ResolutionError(annot, value)

        for var, val in zip(annot.params_type, value.params_type):
            result = self.solve(var, val)
            if not result:
                return result

        if not annot.return_type or not value.return_type:
            return Ok

        result = self.solve(annot.return_type, value.return_type)
        if not result:
            return result

        return Ok

    def solve_Union(self, annot: types.Union, value: types.TypeAttribute):
        for typ in annot.types:
            result = self.solve(typ, value)
            if result:
                return Ok
        return ResolutionError(annot, value)



from abc import ABC, ABCMeta, abstractmethod
from typing import TYPE_CHECKING, TypeVar, ClassVar, Optional
from dataclasses import field, dataclass

from typing_extensions import Self

from kirin.print import Printable
from kirin.ir.traits import Trait
from kirin.lattice.abc import LatticeMeta, SingletonMeta

if TYPE_CHECKING:
    from kirin.ir.dialect import Dialect


class AttributeMeta(ABCMeta):
    """Metaclass for attributes."""

    pass


class LatticeAttributeMeta(LatticeMeta, AttributeMeta):
    """Metaclass for lattice attributes."""

    pass


class SingletonLatticeAttributeMeta(LatticeAttributeMeta, SingletonMeta):
    """Metaclass for singleton lattice attributes."""

    pass


@dataclass(eq=False)
class Attribute(ABC, Printable, metaclass=AttributeMeta):
    """ABC for compile-time values. All attributes are hashable
    and thus need to implement the `__hash__` method.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    dialect: ClassVar[Optional["Dialect"]] = field(default=None, init=False, repr=False)
    """Dialect of the attribute. (default: None)"""
    name: ClassVar[str] = field(init=False, repr=False)
    """Name of the attribute in printing and other text format."""
    traits: ClassVar[frozenset[Trait["Attribute"]]] = field(
        default=frozenset(), init=False, repr=False
    )
    """Set of Attribute traits."""

    @abstractmethod
    def __hash__(self) -> int: ...

    @abstractmethod
    def __eq__(self, value: object) -> bool: ...

    def is_structurally_equal(self, other: Self, context: dict | None = None) -> bool:
        return self == other

    @classmethod
    def has_trait(cls, trait_type: type[Trait["Attribute"]]) -> bool:
        """Check if the Statement has a specific trait.

        Args:
            trait_type (type[Trait]): The type of trait to check for.

        Returns:
            bool: True if the class has the specified trait, False otherwise.
        """
        for trait in cls.traits:
            if isinstance(trait, trait_type):
                return True
        return False

    TraitType = TypeVar("TraitType", bound=Trait["Attribute"])

    def get_trait(self, trait: type[TraitType]) -> Optional[TraitType]:
        """Get the trait of the attribute.

        Args:
            trait (type[Trait]): the trait to get

        Returns:
            Optional[Trait]: the trait if found, None otherwise
        """
        for t in self.traits:
            if isinstance(t, trait):
                return t

        return None



import random

import kirin.prelude
from kirin.dialects import ilist, random as kirin_random


def test_random():

    random.seed(12)
    expected = [random.random() for _ in range(4)]

    @kirin.prelude.basic
    def rnd_main() -> ilist.IList:
        kirin_random.seed(12)
        out = []
        for i in range(4):
            # for i in [1,1,2,3]:  # Same result with this line instead
            out = out + [kirin_random.random()]
        return out

    out: ilist.IList = rnd_main()

    assert out.data == expected


def test_randint():

    random.seed(12)
    expected = [random.randint(i, 10) for i in range(4)]

    @kirin.prelude.basic
    def rndint_main() -> ilist.IList:
        kirin_random.seed(12)
        out = []
        for i in range(4):
            out = out + [kirin_random.randint(i, 10)]
        return out

    out: ilist.IList = rndint_main()

    assert out.data == expected


def test_uniform():

    random.seed(12)
    expected = [random.uniform(i, 10) for i in range(4)]

    @kirin.prelude.basic
    def rnduniform_main() -> ilist.IList:
        kirin_random.seed(12)
        out = []
        for i in range(4):
            out = out + [kirin_random.uniform(i, 10)]
        return out

    out: ilist.IList = rnduniform_main()

    assert out.data == expected
